from dna_features_viewer import BiopythonTranslator

from .tools import annotate_enzymes_sites


def plot_sequence_sites(
    sequence,
    enzymes_names,
    forbidden_enzymes=(),
    unique_sites=True,
    ax=None,
    figure_width=18,
    annotate_inline=True,
):
    """Plot the location of sites in the sequence.

    Non-unique and forbidden sites can be highlighted in red.

    Parameters
    ----------
    sequence: str
        The sequence of interest. ATGC string.

    enzymes_names: list
        List of names of the enzymes to plot.

    forbidden_enzymes: list
        The sites of these enzymes will also be plotted, but with a red background.

    unique_sites: bool
        If true, for each enzyme in enzyme_name with more than one site in the
        sequence, these will be plotted on a red background.

    ax: matplotlib.axes.Axes, optional
        Matplotlib ax on which to draw the figure. If none is provided a new figure
        is created and the ax is returned at the end.

    figure_width: float, optional
        Width of the figure if no ax is provided and a new figure is returned.

    annotate_inline: bool, optional
        If True, the enzyme names will be written inside the annotations when possible,
        instead of above.
    """

    record = annotate_enzymes_sites(
        sequence,
        enzymes_names,
        forbidden_enzymes=forbidden_enzymes,
        unique_sites=unique_sites,
    )
    default_props = dict(
        thickness=10,
        box_color=None,
        fontdict=dict(family="Impact", fontsize=7, color="black", weight="normal"),
    )
    translator = BiopythonTranslator(features_properties=lambda f: default_props)
    graphic_record = translator.translate_record(record)
    graphic_record.labels_spacing = 1
    ax, _ = graphic_record.plot(
        figure_width=figure_width, annotate_inline=annotate_inline, ax=ax
    )
    return ax



# __all__ = []

from .stacked_sites_array import stacked_sites_array
from .tools import annotate_enzymes_sites, write_record
from .plotting import plot_sequence_sites

from .version import __version__



from copy import deepcopy
from Bio import Restriction, SeqIO

try:
    # Biopython <1.78
    from Bio.Alphabet import DNAAlphabet

    has_dna_alphabet = True
except ImportError:
    # Biopython >=1.78
    has_dna_alphabet = False
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.SeqFeature import SeqFeature, FeatureLocation
from dnachisel import DnaNotationPattern


def complement(dna_sequence):
    """Return the complement of the DNA sequence.

    For instance ``complement("ATGCCG")`` returns ``"TACGGC"``.

    Uses Biopython for speed.
    """
    return str(Seq(dna_sequence).complement())


def reverse_complement(sequence):
    """Return the reverse complement of the DNA sequence.

    For instance ``complement("ATGCCG")`` returns ``"GCCGTA"``.

    Uses Biopython for speed.
    """
    return complement(sequence)[::-1]


def annotate_record(
    seqrecord, location="full", feature_type="misc_feature", margin=0, **qualifiers
):
    """Add a feature to a Biopython SeqRecord.

    Parameters
    ----------
    seqrecord : SeqRecord
        The Biopython SeqRecord to be annotated.

    location : tuple
        Either (start, end) or (start, end, strand). Strand defaults to +1.

    feature_type : str
        The type associated with the feature.

    margin : int
        Number of extra bases added on each side of the given location.

    qualifiers : dict
        Dictionary that will be the SeqRecord feature's `qualifiers` attribute.
    """
    if location == "full":
        location = (margin, len(seqrecord) - margin)

    strand = location[2] if len(location) == 3 else 1
    seqrecord.features.append(
        SeqFeature(
            FeatureLocation(location[0], location[1], strand),
            qualifiers=qualifiers,
            type=feature_type,
        )
    )


def sequence_to_biopython_record(
    sequence, id="<unknown id>", name="<unknown name>", features=()
):
    """Return a SeqRecord of the sequence, ready to be Genbanked."""
    if has_dna_alphabet:  # Biopython <1.78
        seq = Seq(sequence, alphabet=DNAAlphabet())
    else:
        seq = Seq(sequence)

    record = SeqRecord(seq, id=id, name=name, features=list(features))
    record.annotations["molecule_type"] = "DNA"

    return record


def enzymes_to_dna_pattern(enzymes_names):
    """Return a dictionary {enzyme_name: DnaNotationPattern()}"""
    return {e: DnaNotationPattern(Restriction.__dict__[e].site) for e in enzymes_names}


def get_enzymes_ATGC_sequences(enzymes_names):
    """Return an ATGC string of the enzyme's recognition site.

    If an enzyme has an ambiguous recognition site (with Ys, Ks, etc.) a valid
    ATGC instance of this representation is returned.
    """
    patterns = enzymes_to_dna_pattern(enzymes_names)
    return {e: p.all_variants()[0] for e, p in patterns.items()}


def find_patterns_matches(seq, patterns=None, enzymes_names=None):
    """Return a dictionary {pattern_name: [sites matches locations]}.

    Parameters
    ----------
    seq : str
        An ATGC string.

    patterns : dict, optional
        A dict of the form {pattern_name: dnachisel.DnaNotationPattern()}.
        A list of enzyme names can be provided instead.

    enzymes_names : list, optional
        A list of enzyme names whose sites will be matched in the sequence.
    """
    if enzymes_names is not None:
        patterns = enzymes_to_dna_pattern(enzymes_names)
    return {e: pattern.find_matches(seq) for e, pattern in patterns.items()}


def annotate_enzymes_sites(
    sequence,
    enzymes_names,
    forbidden_enzymes=(),
    unique_sites=True,
    valid_color="#ccccff",
    invalid_color="#ffcccc",
):
    """Create a record with annotations for cut sites, invalid sites in red.

    Parameters
    ----------
    sequence : str
        An ATGC sequence.

    enzymes_names : list
        List of enzyme names supposed to be in the sequence.

    forbidden_enzymes : list
        List of enzyme names not supposed to be in the sequence. These will
        appear in red.

    unique_sites : bool
        If this is True, then any enzyme site from enzymes_names which appears
        more than once will be labelled in red.

    valid_color : str, optional
        Color for valid sites, by default light blue.

    invalid_color : str, optional
        Color for invalid sites, by default light red.
    """
    record = sequence_to_biopython_record(sequence)
    matches = find_patterns_matches(sequence.upper(), enzymes_names=enzymes_names)
    for enz_name, site_matches in matches.items():
        if unique_sites and len(site_matches) > 1:
            color = invalid_color
        else:
            color = valid_color
        for m in site_matches:
            annotate_record(record, m.to_tuple(), label=enz_name, color=color)
    if len(forbidden_enzymes):
        matches = find_patterns_matches(sequence, enzymes_names=forbidden_enzymes)
        for enz_name, site_matches in matches.items():
            for m in site_matches:
                annotate_record(record, m.to_tuple(), label=enz_name, color="red")
    return record


def write_record(record, target, fmt="genbank"):
    """Write a record as Genbank, FASTA, etc. via Biopython, with fixes.

    This will reduce the record's name to 20 characters or less to avoid
    Biopython throwing an error.
    """
    record = deepcopy(record)
    record.name = record.name[:20]
    if has_dna_alphabet:  # Biopython <1.78
        if str(record.seq.alphabet.__class__.__name__) != "DNAAlphabet":
            record.seq.alphabet = DNAAlphabet()
    record.annotations["molecule_type"] = "DNA"
    if hasattr(target, "open"):
        target = target.open("w")
    SeqIO.write(record, target, fmt)



from zymp import (
    stacked_sites_array,
    plot_sequence_sites,
    annotate_enzymes_sites,
    write_record,
)

enzymes_names = [
    "AccI",
    "AclI",
    "AflII",
    "AflIII",
    "AgeI",
    "ApaLI",
    "AseI",
    "AvaI",
    "BamHI",
    "BanII",
    "BlnI",
    "BmtI",
    "BsmI",
    "BssHII",
    "DdeI",
    "DraI",
    "Eco47III",
    "EcoRI",
    "EcoRV",
    "HindII",
    "HindIII",
    "HinfI",
    "HpaI",
    "KpnI",
    "MfeI",
    "MluI",
    "MspA1I",
    "MunI",
    "NaeI",
    "NcoI",
    "NdeI",
    "NheI",
    "NotI",
    "NsiI",
    "NspI",
    "PstI",
    "PvuI",
    "PvuII",
    "SacI",
    "SacII",
    "SalI",
    "ScaI",
    "SfaNI",
    "SnaBI",
    "SpeI",
    "SphI",
    "SspI",
    "StyI",
    "VspI",
    "XhoI",
    "XmaI",
    "ZraI",
]

forbidden_enzymes = ["BsmBI", "BsaI"]

sequence, enzymes_in_sequence, enzymes_not_in_sequence = stacked_sites_array(
    enzymes_names, forbidden_enzymes=forbidden_enzymes, tries=100
)

print(
    "Sequence length:",
    len(sequence),
    "\nRestriction sites:",
    len(enzymes_in_sequence),
    "\nSites not included: ",
    enzymes_not_in_sequence,
)

# PLOT A SUMMARY
ax = plot_sequence_sites(sequence, enzymes_names)
ax.figure.savefig("stacked_array.pdf", bbox_inches="tight")

# WRITE THE SEQUENCE AND SITE ANNOTATIONS AS A RECORD
record = annotate_enzymes_sites(
    sequence, enzymes_names, forbidden_enzymes=forbidden_enzymes
)
write_record(record, "stacked_site_array.gb")



import os
import matplotlib

matplotlib.use("Agg")

from zymp import (
    stacked_sites_array,
    plot_sequence_sites,
    annotate_enzymes_sites,
    write_record,
)

enzymes_names = [
    "AccI",
    "AclI",
    "AflII",
    "AflIII",
    "AgeI",
    "ApaLI",
    "AseI",
    "AvaI",
    "BamHI",
    "BanII",
    "BlnI",
    "BmtI",
    "BsmI",
    "BssHII",
    "DdeI",
    "DraI",
    "Eco47III",
    "EcoRI",
    "EcoRV",
    "HindII",
    "HindIII",
    "HinfI",
    "HpaI",
    "KpnI",
    "MfeI",
    "MluI",
    "MspA1I",
    "MunI",
    "NaeI",
    "NcoI",
    "NdeI",
    "NheI",
    "NotI",
    "NsiI",
    "NspI",
    "PstI",
    "PvuI",
    "PvuII",
    "SacI",
    "SacII",
    "SalI",
    "ScaI",
    "SfaNI",
    "SnaBI",
    "SpeI",
    "SphI",
    "SspI",
    "StyI",
    "VspI",
    "XhoI",
    "XmaI",
    "ZraI",
]

forbidden_enzymes = ["BsmBI", "BsaI"]


def test_basic_example(tmpdir):

    seq, sites_in_seq, leftover = stacked_sites_array(
        enzymes_names, forbidden_enzymes=forbidden_enzymes, tries=200
    )
    assert len(seq) < 170
    assert len(sites_in_seq) > 43

    # PLOT A SUMMARY
    ax = plot_sequence_sites(seq, enzymes_names)

    # WRITE THE SEQUENCE AND SITE ANNOTATIONS AS A RECORD
    record = annotate_enzymes_sites(
        seq, enzymes_names, forbidden_enzymes=forbidden_enzymes
    )
    write_record(record, os.path.join(str(tmpdir), "test.gb"))


def test_basic_example_with_condition(tmpdir):
    enzymes_names = [
        "AccI",
        "AclI",
        "AflII",
        "AflIII",
        "AgeI",
        "ApaLI",
        "AseI",
        "AvaI",
        "XhoI",
        "SacII",
        "XmaI",
    ]

    for i in range(3):
        for e in ["XhoI", "XmaI"]:

            def success_condition(seq, sites_in_seq, leftover):
                return {e}.issubset(sites_in_seq)

            seq, sites_in_seq, leftover = stacked_sites_array(
                enzymes_names,
                forbidden_enzymes=forbidden_enzymes,
                tries=500,
                success_condition=success_condition,
            )

            assert {e}.issubset(sites_in_seq)



import numpy as np
from dnachisel.biotools import reverse_complement
from Bio import Restriction
import proglog

from .tools import (
    enzymes_to_dna_pattern,
    get_enzymes_ATGC_sequences,
    find_patterns_matches,
    reverse_complement,
)


def sites_difference(site1, site2):
    """Return minimal sequence of nucleotides that should be added at the end
    of site1 to make site2 appear."""
    for i in range(len(site2), -1, -1):
        if site2[:i] == site1[-i:]:
            return site2[i:]
    return site2


def _enzymes_names_to_distances_graph(enzymes_names):
    enzymes_names = enzymes_names[:]
    np.random.shuffle(enzymes_names)
    enzymes_sites = get_enzymes_ATGC_sequences(enzymes_names)
    patterns = enzymes_to_dna_pattern(enzymes_names)
    core_enzymes = {}
    for e1 in list(enzymes_names):
        site1 = enzymes_sites[e1]
        rev_site1 = reverse_complement(site1)
        for e2 in list(enzymes_names):
            if e1 == e2:
                continue
            site2 = enzymes_sites[e2]
            pattern2 = patterns[e2]
            if any(
                [
                    site1 == site2,
                    site2 in site1,
                    rev_site1 == site2,
                    site2 in rev_site1,
                    len(pattern2.find_matches(site1)),
                    len(pattern2.find_matches(rev_site1)),
                ]
            ):
                if e1 not in core_enzymes:
                    core_enzymes[e1] = []
                if e2 not in core_enzymes[e1]:
                    core_enzymes[e1].append(e2)
    graph = {}
    for e1 in enzymes_names:
        site1 = enzymes_sites[e1]
        for e2 in enzymes_names:
            if e1 == e2:
                continue
            site2 = enzymes_sites[e2]
            diff = sites_difference(site1, site2)
            graph[(e1, e2)] = dict(diff=diff, dist=len(diff))
    return graph, enzymes_sites, core_enzymes


def _enzyme_path_to_sequence(path, graph, enzymes_sites):
    """Convert a path of successive enzymes into a sequence."""
    return "".join(
        [enzymes_sites[path[0]]]
        + [graph[(n1, n2)]["diff"] for n1, n2 in zip(path, path[1:])]
    )


class NoSuitableSequenceFound(Exception):
    pass


def stacked_sites_array(
    enzymes_names,
    forbidden_enzymes=(),
    unique_sites=True,
    tries=10,
    logger="bar",
    success_condition=None,
):
    """Generate a sequence.

    Parameters
    ----------
    enzymes_names: list
        Names of enzyme sites which the algorithm should try to include in the
        sequence.

    forbidden_enzymes: list
        List of enzymes names whose sites should not be in the sequence.

    unique_sites: bool
        If True, all the enzymes in enzymes_names will have no more than one site
        in the final sequence.

    tries: int
        Number of tries. The sequence returned is the one with the best score,
        i.e., with the least enzymes left over, or the shortest sequence
        in case of equality.

    success_condition: function
        A function evaluated at the end of each try to validate the designed sequence.
        Should be of the form ``f(seq, inseq, notinseq) => True/False`` where ``seq``
        is the final sequence, ``inseq`` the set of enzymes in the sequence,
        and ``notinseq`` the set of enzymes not in the sequence.

    logger: {'bar', None} or Proglog progress logger, optional
        Either "bar" for a progress bar, None for no progress logger,
        or any Proglog progress logger.

    Returns
    -------
    tuple
        A triplet (sequence, enzymes_in_sequence, enzymes_not_in_sequence).
    """

    enzymes_names = sorted(set(enzymes_names))
    logger = proglog.default_bar_logger(logger, min_time_interval=0.2)
    patterns = enzymes_to_dna_pattern(list(enzymes_names) + list(forbidden_enzymes))
    graph, enzymes_sites, core_enzymes = _enzymes_names_to_distances_graph(
        enzymes_names
    )

    def one_try():
        pickable_enzymes_not_in_seq = set(enzymes_names)
        enzymes_not_in_seq = set(enzymes_names)
        enzymes_in_seq = set()

        def add_enzyme(e, as_core=True):
            if e not in enzymes_in_seq:
                if e in pickable_enzymes_not_in_seq:
                    pickable_enzymes_not_in_seq.remove(e)
                enzymes_not_in_seq.remove(e)
                enzymes_in_seq.add(e)
                if as_core:
                    for enz in core_enzymes.get(e, []):
                        add_enzyme(enz, as_core=False)

        list_pickable_enzymes_not_in_seq = list(pickable_enzymes_not_in_seq)
        path = [
            list_pickable_enzymes_not_in_seq[
                np.random.randint(len(list_pickable_enzymes_not_in_seq))
            ]
        ]
        seq = _enzyme_path_to_sequence(path, graph, enzymes_sites)
        add_enzyme(path[-1])
        while len(pickable_enzymes_not_in_seq):
            nodes_distances = sorted(
                [
                    (graph[(path[-1], n)]["dist"], n)
                    for n in pickable_enzymes_not_in_seq
                ],
                key=lambda dn: (dn[0], np.random.rand()),
            )
            choices = [n for d, n in nodes_distances]
            for new_node in choices:
                new_path = path + [new_node]
                new_seq = _enzyme_path_to_sequence(new_path, graph, enzymes_sites)
                new_diff = len(new_seq) - len(seq)
                end_seq = new_seq[-(6 + new_diff) :]
                matches = find_patterns_matches(end_seq, patterns)
                if unique_sites:
                    new_matches = {
                        e: [m for m in matches[e] if m.end > len(end_seq) - new_diff]
                        for e in matches
                    }
                    forbidden = list(enzymes_in_seq) + list(forbidden_enzymes)
                    if any([len(new_matches[e]) for e in forbidden]):
                        continue
                    if any([len(new_matches[e]) > 1 for e in enzymes_not_in_seq]):
                        continue
                seq = new_seq
                path = new_path
                add_enzyme(new_node)
                for e in list(enzymes_not_in_seq):
                    if len(matches[e]) > 0:
                        add_enzyme(e)
                break
            else:
                return seq, enzymes_in_seq, enzymes_not_in_seq
        return seq, enzymes_in_seq, enzymes_not_in_seq

    failure_score = (10000, 10000)

    def score(seq, in_seq, leftover):
        if unique_sites or len(forbidden_enzymes):
            matches = find_patterns_matches(seq, patterns)
            if unique_sites and any([len(matches[e]) != 1 for e in in_seq]):
                return failure_score
            if any([len(matches[e]) for e in forbidden_enzymes]):
                return failure_score
        if success_condition is not None:
            if not success_condition(seq, in_seq, leftover):
                return failure_score
        return (len(leftover), len(seq))

    current_best = seq, in_seq, leftover = one_try()
    current_score = score(seq, in_seq, leftover)
    for _ in logger.iter_bar(tries=range(tries)):
        new_result = seq, in_seq, leftover = one_try()
        new_score = score(seq, in_seq, leftover)
        if new_score < current_score:
            current_score = new_score
            current_best = new_result
    if current_score == failure_score:
        raise NoSuitableSequenceFound("Could not find a suitable sequence.")
    return current_best
